require "Window"
 
local Precog_Range = {}
 
local WrongInterruptBaseSpellIds = {
	[19190] = true -- Esper's Fade Out
}

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function Precog_Range:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

	self.tUnits = {}

	self.version = "0.1.3"
	self.author = "Lore Walker"
	self.strAddonName = "Precog_Range"
	self.strDisplayName = "Precog_Range"
	self.tX = 0
	self.tY = 0
	self.restored = false
	self.click = false
	self.frameCount = 0
	self.tAnchorOffsets = nil
	
	self.tSettings = {}
	self.tDefaults = {
	    _version = 3,
		range = 30,
		iconSize = 12,
		visible = true,
		frameRate = 2,		-- Draw frames at this rate
		tMarkers = {
			FriendlyAssault = 	{ crObject = "classColor", strIcon = "Precog_RangeSprites:PrecogTri"},
			FriendlyHeal =		{ crObject = "classColor", strIcon = "Precog_RangeSprites:PrecogPlus"},
			FriendlyTank =		{ crObject = "classColor", strIcon = "Precog_RangeSprites:PrecogShield"},
			HostileAssault = 	{ crObject = "classColor", strIcon = "Precog_RangeSprites:PrecogX"},
			HostileHeal = 		{ crObject = "classColor", strIcon = "Precog_RangeSprites:PrecogX"}, -- Can't determine through AP/SP
			HostileTank =		{ crObject = "classColor", strIcon = "Precog_RangeSprites:PrecogX"}, -- Can't determine through AP/SP
			Hostile = 			{ crObject = "FF0000", strIcon = "Precog_RangeSprites:PrecogX" },
			Neutral = 			{ crObject = "FFCC00", strIcon = "Precog_RangeSprites:PrecogX" },
		},
		classSupport = {
			[GameLib.CodeEnumClass.Warrior] =		"Tank",
			[GameLib.CodeEnumClass.Engineer] =		"Tank",
			[GameLib.CodeEnumClass.Esper] =			"Heal",
			[GameLib.CodeEnumClass.Medic] =			"Heal",
			[GameLib.CodeEnumClass.Stalker] =		"Tank",
			[GameLib.CodeEnumClass.Spellslinger] =	"Heal",		
		},
		classColors = {
			[GameLib.CodeEnumClass.Warrior] =		"F54F4F",
			[GameLib.CodeEnumClass.Engineer] =		"EFAB48",
			[GameLib.CodeEnumClass.Esper] =			"1591DB",
			[GameLib.CodeEnumClass.Medic] =			"FFE757",
			[GameLib.CodeEnumClass.Stalker] =		"D23EF4",
			[GameLib.CodeEnumClass.Spellslinger] =	"98C723",
		},
		raidMarkers = {
			"Icon_Windows_UI_CRB_Marker_Bomb",
			"Icon_Windows_UI_CRB_Marker_Ghost",
			"Icon_Windows_UI_CRB_Marker_Mask",
			"Icon_Windows_UI_CRB_Marker_Octopus",
			"Icon_Windows_UI_CRB_Marker_Pig",
			"Icon_Windows_UI_CRB_Marker_Chicken",
			"Icon_Windows_UI_CRB_Marker_Toaster",
			"Icon_Windows_UI_CRB_Marker_UFO",
		},
		interruptFlashDuration = 2.5,
		dispelFlashDuration = 2.5,
	}

	return o
end

function Precog_Range:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)
	Apollo.RegisterEventHandler("UnitGibbed", "OnUnitDestroyed", self)
	Apollo.RegisterEventHandler("UnitActivationTypeChanged", "OnUnitChanged", self)
	Apollo.RegisterEventHandler("UnitMiniMapMarkerChanged", "OnUnitChanged", self)
	Apollo.RegisterEventHandler("CombatLogCCState", "OnCombatLogCCState", self)
	Apollo.RegisterEventHandler("CombatLogVitalModifier", "OnCombatLogVitalModifier", self)
	Apollo.RegisterEventHandler("CombatLogDispel", "OnCombatLogDispel", self)	
end
 
-----------------------------------------------------------------------------------------------
-- Precog_Range OnLoad
-----------------------------------------------------------------------------------------------
function Precog_Range:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("Precog_Range.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- Precog_Range OnDocLoaded
-----------------------------------------------------------------------------------------------
function Precog_Range:OnDocLoaded()
	if self.xmlDoc == nil or not self.xmlDoc:IsLoaded() then return end
	
	Apollo.LoadSprites("Precog_RangeSprites.xml")
	
	self.wndMain = Apollo.LoadForm(self.xmlDoc, "Precog_RangeFinderContainer", "FixedHudStratumLow", self)
	self.wndCanvas = self.wndMain:FindChild("Precog_RangeFinder")
	self.txtTargets = self.wndMain:FindChild("Targets")
	self.txtTargetRange = self.wndMain:FindChild("TargetRange")
	self.txtRange = self.wndMain:FindChild("Range")
	self.wndRange = self.wndMain:FindChild("RangeIndicator")
	
	-- if we didn't restore any data, set defaults
	if (not self.restored) then
		self.tSettings = deepcopy(self.tDefaults)
	end
	
	-- load prior window positioning
	if (self.tAnchorOffsets) then
		self.wndMain:SetAnchorOffsets(unpack(self.tAnchorOffsets))
		self:OnWindowMove(self.wndMain, self.wndMain, 0, 0, 0, 0)
	end
	self.wndMain:Show(self.tSettings.visible, true)
	Apollo.RegisterSlashCommand("range", "OnRange", self)
	
	Apollo.RegisterEventHandler("NextFrame", "UpdateAllUnits", self)
end

function Precog_Range:OnRange(cmd, arg)
    if (arg ~= "") then
		tArgs = {}
		for sArg in arg:gmatch("%w+") do table.insert(tArgs, sArg) end

		if (tArgs[1] == "frameSkip") then
			newRate = tonumber(tArgs[2])
			if (newRate) then
				self.tSettings.frameRate = newRate + 1
			end
		elseif (tArgs[1] == "iconSize") then
			newSize = tonumber(tArgs[2])
			if (newSize) then
				self.tSettings.iconSize = newSize
			end
		else
			newRange = tonumber(tArgs[1])
			if (newRange) then
				self.tSettings.range = newRange
			end
			self.tSettings.visible = true
		
			if (self.tSettings.range >= 128) then
				Print ("Warning: Unit updates and targetting are only reliable for ranges below 128m.")
			end
		end
	else
		self.tSettings.visible = not self.tSettings.visible
	end
	
	self.wndMain:Show(self.tSettings.visible, true)
end

function Precog_Range:OnUnitCreated(unit)
	if (unit == nil) or (not unit:IsValid()) then return end

	self.tUnits[unit:GetId()] = { unit = unit, dispelTime = nil, interruptTime = nil, CCState = {} }
end

function Precog_Range:OnUnitDestroyed(unit)
	if (unit == nil) then return end
	
	self.tUnits[unit:GetId()] = nil
end

function Precog_Range:OnUnitChanged(unit, eType)
	if (unit == nil) then return end
	
	if (not unit:IsValid()) or unit:IsDead() then
		self.tUnits[unit:GetId()] = nil
	else
		self.tUnits[unit:GetId()] = { unit = unit, dispelTime = nil, interruptTime = nil, CCState = {} }
	end
end

function Precog_Range:UpdateAllUnits()
    if (self.wndMain == nil) or (GameLib.GetPlayerUnit() == nil) then return end
	
	self:HandleUnits()
end

function round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function Precog_Range:HandleUnits()
	-- Don't process when the frame isn't showing!
    local player = GameLib.GetPlayerUnit()
    if (self.wndMain == nil) or (player == nil) or (self.tSettings.visible == false) then return end
	
	self.frameCount = self.frameCount + 1
	if (self.frameCount % self.tSettings.frameRate == 0) then
		self.frameCount = 0
	else
		-- only draw at the desired frame rate
		return
	end

	-- Mini-map calculations
	local tTypes = {}
	local eDisposition = 0
	local tInfo = {}
	local uPos = {}
	local uVec = {}
	local dPos = {}
	local nMarkerId = 0
	local classId = 0
	local dx = 0
	local dy = 0
	local dOff = {0, 0, 0, 0}
	local sqrDist = 0
	local sqrDrawRange = math.pow(self.tSettings.range * 2, 2)
	local sqrRange = math.pow(self.tSettings.range, 2)	

	-- Setup tooltip information
	local xml = nil
	local tCount = 0
	local nCount = 0
	
	-- Grab target for highlighting and target marker icons
	local unit = nil
	local targetId = 0
	local playerId = player:GetId()
	self.tUnits[playerId] = nil
	if GameLib.GetTargetUnit() then
		targetId = GameLib.GetTargetUnit():GetId()
	end

	-- Constants used to calculate unit's position with the range frame
	local wndOff = self.wndCanvas:GetWidth() / 2
	local iconOff = self.tSettings.iconSize / 2
	local pPos = player:GetPosition() 
	local pVec = Vector3.New(pPos.x, pPos.y, pPos.z)
	local ppm = self.wndRange:GetWidth() / (self.tSettings.range * 2) -- Pixels/Meter

	local uHeading
    local pHeading = player:GetHeading()
    if (pHeading < 0) then
      pHeading = pHeading * -1
    else
      pHeading = 2 * math.pi - pHeading
    end
	local cosHeading = math.cos(pHeading)
	local sinHeading = math.sin(pHeading)

	-- Reset all current data to draw the next frame
	self.wndCanvas:DestroyAllPixies()
	self.wndCanvas:SetTooltipDoc(nil)
	self.txtRange:SetText(self.tSettings.range .. "m")
	self.txtTargets:SetText("")
	self.txtTargetRange:SetText("")
	
	-- Draw the icon indicating the player's position and facing
	self.wndCanvas:AddPixie({
		bLine = false,
		strSprite = "ClientSprites:MiniMapObjectEdge",
		--strSprite = "Crafting_CoordSprites:sprCoord_Direction_N_Blue",
		cr = {a=1,r=1.00,g=1.00,b=1.00},
		loc = {
			fPoints = {0,0,0,0},
			nOffsets = {wndOff - 12, wndOff - 12, 
						wndOff + 12, wndOff + 12}
		}
	})
	
	-- Draw North
	uVec = Vector3.New(pPos.x,pPos.y,pPos.z - self.tSettings.range)
	dPos = (pVec - uVec)
	-- Rotate based on player facing
	dx = ppm * ((dPos.x * cosHeading) - (-dPos.z * sinHeading))
	dy = ppm * ((dPos.x * sinHeading) + (-dPos.z * cosHeading))
	
	dOff = {-7 *dx/8 + wndOff, 7*dy/8 + wndOff, 
			-dx + wndOff, dy + wndOff}
	self.wndCanvas:AddPixie({
		bLine = true,
		fWidth = math.max(2, self.wndRange:GetWidth() / 40),
		cr = {a=0.5,r=1.00,g=0.00,b=0.00},
		loc = {
			fPoints = {0,0,0,0},
			nOffsets = dOff
		}
	})
	
	-- Draw South
	uVec = Vector3.New(pPos.x,pPos.y,pPos.z + self.tSettings.range)
	dPos = (pVec - uVec)
	-- Rotate based on player facing
	dx = ppm * ((dPos.x * cosHeading) - (-dPos.z * sinHeading))
	dy = ppm * ((dPos.x * sinHeading) + (-dPos.z * cosHeading))
	
	dOff = {-7 *dx/8 + wndOff, 7*dy/8 + wndOff, 
			-dx + wndOff, dy + wndOff}
	self.wndCanvas:AddPixie({
		bLine = true,
		fWidth = math.max(2, self.wndRange:GetWidth() / 40),
		cr = {a=0.5,r=1.00,g=1.00,b=1.00},
		loc = {
			fPoints = {0,0,0,0},
			nOffsets = dOff
		}
	})
	
	-- Draw East
	uVec = Vector3.New(pPos.x - self.tSettings.range,pPos.y,pPos.z)
	dPos = (pVec - uVec)
	-- Rotate based on player facing
	dx = ppm * ((dPos.x * cosHeading) - (-dPos.z * sinHeading))
	dy = ppm * ((dPos.x * sinHeading) + (-dPos.z * cosHeading))
	
	dOff = {-7 *dx/8 + wndOff, 7*dy/8 + wndOff, 
			-dx + wndOff, dy + wndOff}
	self.wndCanvas:AddPixie({
		bLine = true,
		fWidth = math.max(2, self.wndRange:GetWidth() / 40),
		cr = {a=0.5,r=1.00,g=1.00,b=1.00},
		loc = {
			fPoints = {0,0,0,0},
			nOffsets = dOff
		}
	})
	
	-- Draw West
	uVec = Vector3.New(pPos.x + self.tSettings.range,pPos.y,pPos.z)
	dPos = (pVec - uVec)
	-- Rotate based on player facing
	dx = ppm * ((dPos.x * cosHeading) - (-dPos.z * sinHeading))
	dy = ppm * ((dPos.x * sinHeading) + (-dPos.z * cosHeading))
	
	dOff = {-7 *dx/8 + wndOff, 7*dy/8 + wndOff, 
			-dx + wndOff, dy + wndOff}
	self.wndCanvas:AddPixie({
		bLine = true,
		fWidth = math.max(2, self.wndRange:GetWidth() / 40),
		cr = {a=0.5,r=1.00,g=1.00,b=1.00},
		loc = {
			fPoints = {0,0,0,0},
			nOffsets = dOff
		}
	})
	
	for unitId, tUnit in pairs(self.tUnits) do
		unit = tUnit.unit
		if (unit:IsValid()) and (not unit:IsDead()) and (unit:ShouldShowNamePlate()) then 
		
			uPos = unit:GetPosition()
			uVec = Vector3.New(uPos.x,uPos.y,uPos.z)
			dPos = (pVec - uVec)
			
			-- Only draw/update units within 2x the range indicator
			sqrDist = math.pow(dPos.x, 2) + math.pow(dPos.y, 2) + math.pow(dPos.z, 2)
			if (sqrDist <= sqrDrawRange) then
				classId = unit:GetClassId()
			
				-- players
				if (unit:GetType() == "Player") then
					eDispotition = unit:GetDispositionTo(player)
					if eDispotition == Unit.CodeEnumDisposition.Hostile then
						tTypes = { "HostileAssault" }
					elseif eDispotition == Unit.CodeEnumDisposition.Friendly then
						if (unit:GetAssaultPower() > unit:GetSupportPower()) then
							tTypes = { "FriendlyAssault" }
						else
							tTypes = { "Friendly" .. self.tSettings.classSupport[classId] }						
						end
					end
				else
					tTypes = unit:GetMiniMapMarkers()
				end
		
				for idx, type in pairs(tTypes) do
					if self.tSettings.tMarkers[type] then
						tMarker = self.tSettings.tMarkers[type]
					
						if tMarker.strIcon then
							tInfo.strIcon = tMarker.strIcon 
						end
	
						if tMarker.crObject then
							tInfo.crObject = tMarker.crObject
						end
						if tMarker.crObject == "classColor" then
							if self.tSettings.classColors[classId] then
								tInfo.crObject = self.tSettings.classColors[classId]
							end
						end
							
						-- Rotate based on player facing
						dx = ppm * ((dPos.x * cosHeading) - (-dPos.z * sinHeading))
						dy = ppm * ((dPos.x * sinHeading) + (-dPos.z * cosHeading))
						
						dOff = {-dx + wndOff - iconOff, dy + wndOff - iconOff, 
								-dx + wndOff + iconOff, dy + wndOff + iconOff}

						-- Modifiers
						if (unitId == targetId) then -- Target
							self.wndCanvas:AddPixie({
								bLine = false,
								strSprite = "ClientSprites:HuddieLevelCircle",							
								cr = "9A" .. tInfo.crObject,
								loc = {
									fPoints = {0,0,0,0},
									nOffsets = {dOff[1] - 2, dOff[2] - 2, 
												dOff[3] + 2, dOff[4] + 2}
								}
							})
							
							-- sqrt is expensive, only calculated for target -- may use cache in the future
							self.txtTargetRange:SetText(round(dPos:Length(), 1).."m") 
						end
						
						nMarkerId = unit:GetTargetMarker() or 0
						if nMarkerId and self.tSettings.raidMarkers[nMarkerId] then -- Raid Marker
							self.wndCanvas:AddPixie({
								bLine = false,
								strSprite = self.tSettings.raidMarkers[nMarkerId],
								cr = "FFFFFFFF",
								loc = {
									fPoints = {0,0,0,0},
									nOffsets = {dOff[1] - 8, dOff[2] - 8,
												dOff[3] + 8, dOff[4] + 8}
								}
							})						
						end						
						
						if (tUnit.dispelTime ~= nil) then
							dTime = os.difftime(os.time(), tUnit.dispelTime)
							if (dTime <= self.tSettings.dispelFlashDuration) then
								self.wndCanvas:AddPixie({
									bLine = false,
									strSprite = "ClientSprites:HuddieLevelCircle",
									cr = {a=(self.tSettings.dispelFlashDuration - dTime)/ self.tSettings.dispelFlashDuration,
									      r=1.00,g=1.00,b=1.00},
									loc = {
										fPoints = {0,0,0,0},
										nOffsets = {dOff[1] - 2, dOff[2] - 2,
													dOff[3] + 2, dOff[4] + 2}
									}
								})
							else
								tUnit.dispelTime = nil
							end
						end

						if (tUnit.interruptTime ~= nil) then
							dTime = os.difftime(os.time(), tUnit.interruptTime)
							if (dTime <= self.tSettings.interruptFlashDuration) then
								self.wndCanvas:AddPixie({
									bLine = false,
									strSprite = "ClientSprites:HuddieLevelCircle",
									cr = {a=(self.tSettings.interruptFlashDuration - dTime)/ self.tSettings.interruptFlashDuration,
									      r=1.00,g=1.00,b=1.00},
									loc = {
										fPoints = {0,0,0,0},
										nOffsets = {dOff[1] - 2, dOff[2] - 2,
													dOff[3] + 2, dOff[4] + 2}
									}
								})
							else
								tUnit.interruptTime = nil
							end
						end
						
						for name, state in pairs(Unit.CodeEnumCCState) do
							if (unit:IsInCCState(state)) then
								ccTime = unit:GetCCStateTimeRemaining(state)
								if (tUnit.CCState[state] == nil) then
									tUnit.CCState[state] = ccTime
								end
								self.wndCanvas:AddPixie({
									bLine = false,
									strSprite = "BasicSprites:WhiteCircle",
									cr = {a=ccTime/tUnit.CCState[state],
									      r=1.00,g=0.00,b=0.75},
									loc = {
										fPoints = {0,0,0,0},
										nOffsets = {dOff[1] - 2, dOff[2] - 2,
													dOff[3] + 2, dOff[4] + 2}
									}
								})
							else
								tUnit.CCState[state] = nil
							end
						end	
						
						-- Draw Unit
						self.wndCanvas:AddPixie({
							bLine = false,
							strSprite = tInfo.strIcon,
							cr = "FF" .. tInfo.crObject,
							loc = {
								fPoints = {0,0,0,0},
								nOffsets = dOff
							}
						})						
					
						-- Track the number of friendly/party units that are in the specified range
						if ((type == "FriendlyAssault") or (type == "FriendlyHeal") or (type == "FriendlyTank")) and (sqrDist <= sqrRange) then
							nCount = nCount + 1
						end
					
						-- Check if we need to setup a tooltip
						if (self.tX <= dOff[3]) and (self.tX >= dOff[1] - 2) and (self.tY <= dOff[4]) and (self.tY >= dOff[2] - 2) then
							tCount = tCount + 1
							
							-- Only waste time creating xml tooltip doc when a tooltip is needed
							if (xml == nil) then
								xml = XmlDoc.new()
								xml:StartTooltip(Tooltip.TooltipWidth)
							end
							xml:AddLine(unit:GetName() .. " (z: " .. round(-dPos.y, 0) .. ")", crWhite, "CRB_InterfaceMedium")
							
							-- If a click occurred, then target the first unit next to the mouse
							if (self.click) and (unitId ~= targetId) then
								GameLib.SetTargetUnit(unit)
								self.click = false
							end
						end
					end
				end
			end
		else
			self.tUnits[unitId] = nil
		end
	end
	
	self.click = false
	
	if (nCount > 0) then
		self.txtTargets:SetText("" .. nCount)
	end
	
	if (tCount > 0) and (xml ~= nil) then
		self.wndMain:SetTooltipDoc(xml)
	else
		self.wndMain:SetTooltipDoc(nil)
	end	
end

-- Tooltips are generated by the frame handler
function Precog_Range:OnGenerateTooltip(wndHandler, wndControl, eType, nX, nY)
    self.tX = nX
	self.tY = nY
end

function Precog_Range:OnMouseMove(wndHandler, wndControl, nX, nY)
    self.tX = nX
	self.tY = nY
end

-- Targetting is done when we build the frame
function Precog_Range:OnClick(wndHandler, wndControl, eMouseButton, nX, nY, bDoubleClick, bStopPropagation)
	if eMouseButton ~= GameLib.CodeEnumInputMouse.Left then return end
    self.click = true
end

function Precog_Range:OnMouseWheel(wndHandler, wndControl, nX, nY, fScrollAmount, bConsumeMouseWheel)
	if Apollo.IsShiftKeyDown() then
		if math.ceil(fScrollAmount) > 0 then
			nScroll = -1/3
		else
			nScroll = 1/3
		end
		if (self.tSettings.range + nScroll > 1) then
			self.tSettings.range = self.tSettings.range + nScroll
		end
	else
		nScroll = -math.ceil(fScrollAmount)
		if (self.tSettings.range + nScroll > 0) then
			self.tSettings.range = math.ceil(self.tSettings.range + nScroll)
		end
	end
end

function Precog_Range:OnWindowMove(wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom)
	local l,t,r,b = self.wndMain:GetAnchorOffsets()
	local length = math.min(r - l, b - t) - 2
	
	self.wndCanvas:SetAnchorOffsets(1, 1, length, length)
end

function Precog_Range:OnCombatLogCCState(e)
	if e.nInterruptArmorHit > 0 and e.unitCaster and not WrongInterruptBaseSpellIds[e.splCallingSpell:GetBaseSpellId()] then
		local tUnit = self.tUnits[e.unitCaster:GetId()]
		if tUnit and tUnit.unit then
			tUnit.interruptTime = os.time()
		end
	end
end

function Precog_Range:OnCombatLogVitalModifier(e)
	if e.eVitalType == GameLibCodeEnumVitalInterruptArmor and e.unitCaster and e.nAmount < 0 then
		local tUnit = self.tUnits[e.unitCaster:GetId()]
		if tUnit and tUnit.unit then
			tUnit.interruptTime = os.time()
		end
	end
end

function Precog_Range:OnCombatLogDispel(e)
	if e.unitCaster and e.nInstancesRemoved > 0 then
		local tUnit = self.tUnits[e.unitCaster:GetId()]
		if tUnit and tUnit.unit then
			tUnit.dispelTime = os.time()
		end
	end
end

-----------------------------------------------------------------------------------------------
-- OnSave / OnRestore
-----------------------------------------------------------------------------------------------
function Precog_Range:OnSave(eLevel)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then 
		return 
  	end
	
	return {
		_settings = deepcopy(self.tSettings),
		_tAnchorOffsets = {self.wndMain:GetAnchorOffsets()}
	}
end

function Precog_Range:OnRestore(eLevel, tData)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then 
		return 
  	end
	
	if tData._settings ~= nil then
		if (tData._settings._version == self.tDefaults._version) then
			self.tSettings = deepcopy(tData._settings)
		else
			-- new version, must reset to defaults
			Print("Precog_Range: New version detected! Loading Defaults...")
			self.tSettings = deepcopy(self.tDefaults)
			
			if (tData._settings.range ~= nil) then
				self.tSettings.range = tData._settings.range
			end

			if (tData._settings.iconSize ~= nil) then
				self.tSettings.iconSize = tData._settings.iconSize
			end

			if (tData._settings.frameRate ~= nil) then
				self.tSettings.frameRate = tData._settings.frameRate
			end				
			
			if (tData._settings.visible ~= nil) then
				self.tSettings.visible = tData._settings.visible
			end
		end
	else
		-- No save data, load defaults
		self.tSettings = deepcopy(self.tDefaults)
	end
	
	if tData._tAnchorOffsets ~= nil then
		self.tAnchorOffsets = deepcopy(tData._tAnchorOffsets)
	end
	
	self.restored = true
end

function deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

-----------------------------------------------------------------------------------------------
-- Precog_Range Instance
-----------------------------------------------------------------------------------------------
local Precog_RangeInst = Precog_Range:new()
Precog_RangeInst:Init()
